import os
import logging
from pathlib import Path
from argparse import ArgumentParser

from src.to_csv import write_measurements_to_csv_files
from src.read_dataset import read_datasets_from_directory
from loggingUtility import set_log_level, add_log_level_to_argument_parser


def create_output_directory(datasetOutputPath: Path) -> bool:
    if os.path.exists(datasetOutputPath):
        if not os.path.isdir(datasetOutputPath):
            logging.error(f"Unable to store measurements in '{datasetOutputPath}'")
            return False
    else:
        logging.debug(f"Create directory {datasetOutputPath}")
        os.makedirs(datasetOutputPath)

    return True


def main() -> None:
    parser = ArgumentParser(description='Transform dataset to C++ header')
    parser.add_argument('datasetFolder', type=str, help='Path to dataset folder')
    parser.add_argument('--outputFolder', type=str, default='./output', help='Path to output folder (default="./output")')
    add_log_level_to_argument_parser(parser)
    args = parser.parse_args()
    set_log_level(args.logLevel)

    outputDir = Path(args.outputFolder)

    datasets = read_datasets_from_directory(args.datasetFolder)
    for dataset in datasets:
        logging.info(f"Process dataset {dataset.name}")
        if not dataset.measurements:
            logging.info(f"Dataset {dataset.name} has no measurements. Skip it.")
            continue

        datasetOutputPath = outputDir/dataset.name
        success = create_output_directory(datasetOutputPath)
        if not success:
            return 1

        measurements = dataset.measurements

        numberOfDigits = len(str(len(measurements)))
        for index, measurementSeries in enumerate(measurements):
            paddedIndex = f"{index:>0{numberOfDigits}}"
            write_measurements_to_csv_files(measurementSeries,
                                            datasetOutputPath/f"measurements_{paddedIndex}.csv",
                                            datasetOutputPath/f"weights_{paddedIndex}.csv")

    return 0


if __name__ == "__main__":
    exit(main())
