from argparse import ArgumentParser
from pathlib import Path
import json
import logging
import re

from loggingUtility import set_log_level, add_log_level_to_argument_parser

from src.read_dataset import read_datasets_from_directory
from src.Dataset import ModelType, Dataset
from src.read_pcomp_model import VariablesDimensionError, FunctionsDimensionError, Function, PCOMPModel, read_pcomp_model, get_pcomp_model_dimensions_from


def find_whole_word(targetString: str, sourceString: str, ignoreCase: bool = True) -> bool:
    if ignoreCase:
        flag = re.IGNORECASE
    else:
        flag: re.RegexFlag = 0

    if re.search(r"\b" + re.escape(targetString) + r"\b", sourceString, flags=flag):
        return True
    
    return False


def any_function_contains_symbol(functions: list[Function], symbol: str) -> bool:
    return any([find_whole_word(symbol, '\n'.join(function.body), ignoreCase=True) for function in functions])


def any_function_contains_symbols(functions: list[Function], symbols: list[str]) -> bool:
    return any([any_function_contains_symbol(functions, symbol) for symbol in symbols])


def get_model_parameters_not_in_ode(pcompModel: PCOMPModel) -> list[str]:
    return [parameter for parameter in pcompModel.variables.modelParameters 
                        if not any_function_contains_symbol(pcompModel.functions.differentialEquation, parameter)]


def get_constraints_status(modelFileName: Path, dataset: Dataset) -> tuple[bool, bool]:
    readingPcompModelWasSuccessfulFlag = True
    readingPcompModelWasNotSuccessfulFlag = False
    
    if dataset.totalNumberOfConstraints == 0:
        logging.info(f"{dataset.name}: Problem has no constraints, so constraints are considered valid.")
        return readingPcompModelWasSuccessfulFlag, True

    try:
        pcompModel = read_pcomp_model(modelFileName, get_pcomp_model_dimensions_from(dataset))
    except VariablesDimensionError as err:
        logging.warning(f"{dataset.name}: {err}. Indicate parsing error and return invalid constraints status.")
        return readingPcompModelWasNotSuccessfulFlag, False
    except FunctionsDimensionError as err:
        logging.warning(f"{dataset.name}: {err}. Indicate parsing error and return invalid constraints status.")
        return readingPcompModelWasNotSuccessfulFlag, False
    
    assert pcompModel.functions.constraints is not None

    if any_function_contains_symbols(pcompModel.functions.constraints, pcompModel.variables.states):
        logging.warning(f"{dataset.name}: Constraints depend on states. Return invalid constraints status.")
        return readingPcompModelWasSuccessfulFlag, False
    
    if any_function_contains_symbols(pcompModel.functions.constraints, [pcompModel.variables.time]):
        logging.warning(f"{dataset.name}: Constraints depend on time. Return invalid constraints status.")
        return readingPcompModelWasSuccessfulFlag, False

    if (pcompModel.variables.concentration is not None
        and any_function_contains_symbols(pcompModel.functions.constraints, [pcompModel.variables.concentration])):
        logging.warning(f"{dataset.name}: Constraints depend on concentration. Return invalid constraints status.")
        return readingPcompModelWasSuccessfulFlag, False

    if any_function_contains_symbols(pcompModel.functions.constraints, get_model_parameters_not_in_ode(pcompModel)):
        logging.warning(f"{dataset.name}: Constraints depend on model parameters that are not in ODE. Return invalid constraints status.")
        return readingPcompModelWasSuccessfulFlag, False

    logging.info(f"{dataset.name}: Constraints are valid.")
    return readingPcompModelWasSuccessfulFlag, True


def read_broken_models() -> list[str]:
    fileContent = Path("broken_models.txt").read_text()
    return [line for line in fileContent.splitlines() if line and not line.startswith("#")]


def main() -> int:
    parser = ArgumentParser(description='Read constraints status from PCOMP file and write to JSON')
    parser.add_argument('datasetFolder', type=str, help='Path to dataset folder')
    parser.add_argument('modelFolder', type=str, help='Path to model folder')
    parser.add_argument('outputFilePath', type=str, help='Path to output file')
    add_log_level_to_argument_parser(parser)
    args = parser.parse_args()
    set_log_level(args.logLevel)

    datasets = read_datasets_from_directory(args.datasetFolder)
    brokenModels = read_broken_models() 

    indicesOfMeasuredStates: dict[str, bool] = {}

    for dataset in datasets:
        if not dataset.modelType == ModelType.OrdinaryDifferentialEquations:
            logging.info(f"Dataset {dataset.name} is not an ODE model. Skip it.")
            continue
        
        if dataset.name in brokenModels:
            logging.info(f"Dataset {dataset.name} is broken. Skip it.")
            continue

        correspondingModeFileName = Path(args.modelFolder) / f"{dataset.name}.fun"

        readingModelSucceeded, status = get_constraints_status(correspondingModeFileName, dataset)
        indicesOfMeasuredStates[dataset.name] = {
            "readingModelSucceeded": readingModelSucceeded,
            "status": status
        }

    outputFolder = Path(args.outputFilePath).parent
    outputFolder.mkdir(parents=True, exist_ok=True)
    
    with Path(args.outputFilePath).open("w") as outfile:
        outfile.write(json.dumps(indicesOfMeasuredStates, indent=4))

    return 0

if __name__ == "__main__":
    exit(main())
