from argparse import ArgumentParser
from pathlib import Path
import json
import logging

from loggingUtility import set_log_level, add_log_level_to_argument_parser

from src.read_dataset import read_datasets_from_directory
from src.Dataset import ModelType, Dataset
from src.read_pcomp_model import VariablesDimensionError, FunctionsDimensionError, read_pcomp_model, get_pcomp_model_dimensions_from


def get_indices_of_measured_states(modelFileName: Path, dataset: Dataset) -> tuple[bool, list[int]]:
    
    try:
        pcompModel = read_pcomp_model(modelFileName, get_pcomp_model_dimensions_from(dataset))
    except VariablesDimensionError as err:
        logging.warning(f"{dataset.name}: {err}. Indicate parsing error and return empty indices of measured states.")
        return False, []
    except FunctionsDimensionError as err:
        logging.warning(f"{dataset.name}: {err}. Indicate parsing error and return empty indices of measured states.")
        return False, []

    readingPcompModelWasSuccessful = True

    indicesOfMeasuredStates =  []
    for function in pcompModel.functions.fittingFunction:
        if len(function.body) != 1:
            logging.warning(f"{dataset.name}: Not a standard fitting function, body has multiple lines. Return empty indices of measured states.")
            return readingPcompModelWasSuccessful, []

        returnStatement = function.body[0].lower().replace(function.name.lower(), "")
        returnStatement = returnStatement.replace("=", "")
        returnStatement = returnStatement.strip()

        try:
            indicesOfMeasuredStates.append([s.lower() for s in pcompModel.variables.states].index(returnStatement.lower()))
        except ValueError:
            logging.warning(f"{dataset.name}: Not a standard fitting function, returns not a single state. Return empty indices of measured states.")
            return readingPcompModelWasSuccessful, []
    
    if indicesOfMeasuredStates != sorted(indicesOfMeasuredStates):
        logging.warning(f"{dataset.name}: Found indices of measured states are not sorted: [{', '.join([f'{index}' for index in indicesOfMeasuredStates])}]. Return empty indices of measured states.")
        return readingPcompModelWasSuccessful, []

    logging.info(f"{dataset.name}: Found indices of measured states: [{', '.join([f'{index}' for index in indicesOfMeasuredStates])}]")
    return readingPcompModelWasSuccessful, indicesOfMeasuredStates


def read_broken_models() -> list[str]:
    fileContent = Path("broken_models.txt").read_text()
    return [line for line in fileContent.splitlines() if line and not line.startswith("#")]


def main() -> int:
    parser = ArgumentParser(description='Read indices of measured states from PCOMP file and write to JSON')
    parser.add_argument('datasetFolder', type=str, help='Path to dataset folder')
    parser.add_argument('modelFolder', type=str, help='Path to model folder')
    parser.add_argument('outputFilePath', type=str, help='Path to output file')
    add_log_level_to_argument_parser(parser)
    args = parser.parse_args()
    set_log_level(args.logLevel)

    datasets = read_datasets_from_directory(args.datasetFolder)
    brokenModels = read_broken_models() 

    indicesOfMeasuredStates: dict[str, tuple[bool, list[int]]] = {}

    for dataset in datasets:
        if not dataset.modelType == ModelType.OrdinaryDifferentialEquations:
            logging.info(f"Dataset {dataset.name} is not an ODE model. Skip it.")
            continue
        
        if dataset.name in brokenModels:
            logging.info(f"Dataset {dataset.name} is broken. Skip it.")
            continue

        correspondingModeFileName = Path(args.modelFolder) / f"{dataset.name}.fun"

        readingModelSucceeded, indices = get_indices_of_measured_states(correspondingModeFileName, dataset)
        indicesOfMeasuredStates[dataset.name] = {
            "readingModelSucceeded": readingModelSucceeded,
            "indices": indices
        }

    outputFolder = Path(args.outputFilePath).parent
    outputFolder.mkdir(parents=True, exist_ok=True)
    
    with Path(args.outputFilePath).open("w") as outfile:
        outfile.write(json.dumps(indicesOfMeasuredStates, indent=4))

    return 0

if __name__ == "__main__":
    exit(main())
