from argparse import ArgumentParser
from pathlib import Path
import json
import logging
import re
from dataclasses import dataclass
from typing import Optional

from loggingUtility import set_log_level, add_log_level_to_argument_parser

from src.read_dataset import read_datasets_from_directory
from src.Dataset import ModelType, Dataset
from src.read_pcomp_model import VariablesDimensionError, FunctionsDimensionError, PCOMPVariables, PCOMPConstant, read_pcomp_model, get_pcomp_model_dimensions_from


def find_whole_word(targetString: str, sourceString: str, ignoreCase: bool = True) -> bool:
    if ignoreCase:
        flag = re.IGNORECASE
    else:
        flag: re.RegexFlag = 0

    if re.search(r"\b" + re.escape(targetString) + r"\b", sourceString, flags=flag):
        return True

    return False


def is_number(string: str) -> bool:
    try:
        float(string)
    except ValueError:
        return False

    return True


def is_constant(string: str, constants: list[PCOMPConstant]) -> bool:
    return string.lower() in [c.name.lower() for c in constants]


def is_model_parameter(string: str, variables: PCOMPVariables) -> bool:
    return string.lower() in [p.lower() for p in variables.modelParameters]


def is_concentration(string: str, variables: PCOMPVariables) -> bool:
    if variables.concentration is not None:
        return string.lower() == variables.concentration.lower()

    return False


@dataclass
class Bounds:
    lower: float
    upper: float

    def to_dict(self) -> dict:
        return {
            "lower": self.lower,
            "upper": self.upper
        }


@dataclass
class InitialStateInfo:
    isParameter: bool
    isFixed: bool
    exactValues: list[float]  # one entry for each measurement file
    bounds: Bounds

    def to_dict(self) -> dict:
        return {
            "isParameter": self.isParameter,
            "isFixed": self.isFixed,
            "exactValues": self.exactValues,
            "bounds": self.bounds.to_dict()
        }

    def __str__(self) -> str:
        return json.dumps(self.to_dict(), indent=2)


def get_initial_state_info(modelFileName: Path, dataset: Dataset) -> tuple[bool, list[InitialStateInfo]]:
    try:
        pcompModel = read_pcomp_model(modelFileName, get_pcomp_model_dimensions_from(dataset))
    except VariablesDimensionError as err:
        logging.warning(f"{dataset.name}: {err}. Return no detected initial state values.")
        return False, []
    except FunctionsDimensionError as err:
        logging.warning(f"{dataset.name}: {err}. Return no detected initial state values.")
        return False, []

    readingPcompModelWasSuccessful = True

    initialStateInfo: list[InitialStateInfo] = []

    for function in pcompModel.functions.initialValues:
        if len(function.body) != 1:
            logging.warning(
                f"{dataset.name}: Not a trivial initial value function, body has multiple lines. Return empty InitialStateInfo.")
            return readingPcompModelWasSuccessful, []

        returnStatement = function.body[0].lower().replace(function.name.lower(), "")
        returnStatement = returnStatement.replace("=", "")
        returnStatement = returnStatement.strip()

        if is_number(returnStatement):
            value = float(returnStatement)
            logging.info(f"Found initial state value to be a fixed number: {value}")
            initialStateInfo.append(
                InitialStateInfo(
                    isParameter=False,
                    isFixed=True,
                    exactValues=[value] * max(dataset.numberOfConcentrations, 1),
                    bounds=Bounds(value, value)
                )
            )
        elif is_constant(returnStatement, pcompModel.constants):
            constantIndex = [c.name.lower() for c in pcompModel.constants].index(returnStatement)
            value = pcompModel.constants[constantIndex].value
            logging.info(f"Found initial state value to be a constant: {value}")
            initialStateInfo.append(
                InitialStateInfo(
                    isParameter=False,
                    isFixed=True,
                    exactValues=[value] * max(dataset.numberOfConcentrations, 1),
                    bounds=Bounds(value, value)
                )
            )

        elif is_model_parameter(returnStatement, pcompModel.variables):
            try:
                modelParameterIndex = [info.name.lower() for info in dataset.modelParameterInfos].index(returnStatement.lower())
                logging.info(
                    f"Found initial state value to be a model parameter of the model: {dataset.modelParameterInfos[modelParameterIndex].name}")
                exactValue = dataset.modelParameterInfos[modelParameterIndex].exactValue
                initialStateInfo.append(
                    InitialStateInfo(
                        isParameter=True,
                        isFixed=False,
                        exactValues=[exactValue] * max(dataset.numberOfConcentrations, 1) if exactValue is not None else [],
                        bounds=Bounds(dataset.modelParameterInfos[modelParameterIndex].lowerBound,
                                      dataset.modelParameterInfos[modelParameterIndex].upperBound)
                    )
                )
            except ValueError as err:
                logging.warning(f"{dataset.name}: Could not find parameter in dataset: {err}")
                return readingPcompModelWasSuccessful, []
        elif is_concentration(returnStatement, pcompModel.variables):
            concentrations = dataset.get_concentration_values()
            initialStateInfo.append(
                InitialStateInfo(
                    isParameter=False,
                    isFixed=False,
                    exactValues=concentrations,
                    bounds=Bounds(min(concentrations), max(concentrations))
                )
            )
        else:
            logging.warning(
                f"{dataset.name}: Not a trivial initial value function, return value is neither a fixed value nor a parameter. Return empty InitialStateInfo.")

            return readingPcompModelWasSuccessful, []

    newline = '\n'
    logging.info(
        f"{dataset.name}: Found initial state info: [\n{f',{newline}'.join([f'{info}' for info in initialStateInfo])}\n]")

    return readingPcompModelWasSuccessful, initialStateInfo


def read_broken_models() -> list[str]:
    fileContent = Path("broken_models.txt").read_text()
    return [line for line in fileContent.splitlines() if line and not line.startswith("#")]


def main() -> int:
    parser = ArgumentParser(description='Read initial state info from PCOMP file and write to JSON')
    parser.add_argument('datasetFolder', type=str, help='Path to dataset folder')
    parser.add_argument('modelFolder', type=str, help='Path to model folder')
    parser.add_argument('outputFilePath', type=str, help='Path to output file')
    add_log_level_to_argument_parser(parser)
    args = parser.parse_args()
    set_log_level(args.logLevel)

    datasets = read_datasets_from_directory(args.datasetFolder)
    brokenModels = read_broken_models()

    initialStateInfoDict: dict[str, object] = {}

    for dataset in datasets:
        if not dataset.modelType == ModelType.OrdinaryDifferentialEquations:
            logging.info(f"Dataset {dataset.name} is not an ODE model. Skip it.")
            continue

        if dataset.name in brokenModels:
            logging.info(f"Dataset {dataset.name} is broken. Skip it.")
            continue

        correspondingModeFileName = Path(args.modelFolder) / f"{dataset.name}.fun"
        readingModelSucceeded, initialStateInfo = get_initial_state_info(correspondingModeFileName, dataset)

        initialStateInfoDict[dataset.name] = {
            "readingModelSucceeded": readingModelSucceeded,
            "initialStateInfo": [info.to_dict() for info in initialStateInfo],
            "numberOfConcentrations": dataset.numberOfConcentrations
        }

    outputFolder = Path(args.outputFilePath).parent
    outputFolder.mkdir(parents=True, exist_ok=True)

    with Path(args.outputFilePath).open("w") as outfile:
        outfile.write(json.dumps(initialStateInfoDict, indent=4))

    return 0


if __name__ == "__main__":
    exit(main())
