import logging
import os
from typing import Iterator, TextIO, List, Tuple

from .Dataset import ModelType, MeasurementType, ConstraintInfo, Measurements, ModelParameterInfo, Dataset


class MeasurementInconsistencyError(Exception):
    def __init__(self, message):
        self.message = message


def is_valid_model_type(modelType: int) -> bool:
    return modelType in [t.value for t in ModelType]


def to_model_type(modelType: int) -> ModelType:
    if not is_valid_model_type(modelType):
        logging.warn(f"Integer {modelType} is not convertible to ModelType")
        return ModelType(0)

    return ModelType(modelType)


def to_measurement_type(measurementTypeAsTring: str) -> MeasurementType:
    lookup = {
        "Simulation (0% error)": MeasurementType.SimulationUniformError_0,
        "Simulation (0.5% error)": MeasurementType.SimulationUniformError_0_5,
        "Simulation (1% error)": MeasurementType.SimulationUniformError_1,
        "Simulation (5% error)": MeasurementType.SimulationUniformError_5,
        "Simulation (10% error)": MeasurementType.SimulationUniformError_10,
        "Simulation (50% error)": MeasurementType.SimulationUniformError_50,
        "Simulation (sigma=1%)": MeasurementType.SimulationNormalError_1,
        "Simulation (sigma=5%)": MeasurementType.SimulationNormalError_5,
        "Experimental": MeasurementType.Experimental,
        "Exact": MeasurementType.Exact,
        "Experimental design": MeasurementType.ExperimentalDesign,
        "none": MeasurementType.Unknown
    }

    if not measurementTypeAsTring in lookup.keys():
        logging.warn(f"String {measurementTypeAsTring} is not convertible to MeasurementType")
        return MeasurementType.Unknown

    return lookup[measurementTypeAsTring]


def read_general_info_to_dataset(linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    lastLineOfGeneralInfo = 8

    for i, line in enumerate(linesIt):
        line = line.rstrip()

        if i in [3, 4, 6, 7]:
            # skip line
            continue
        elif i == 0:
            splittedLine = line.split("\\")
            assert len(splittedLine) == 2
            dataset.name = splittedLine[1]
        elif i == 1:
            splittedLine = line.split()
            assert len(splittedLine) == 2
            dataset.modelName = splittedLine[0]

            modelTypeAsInt = int(splittedLine[1])
            dataset.modelType = to_model_type(modelTypeAsInt)
        elif i == 2:
            dataset.description = line
        elif i == 5:
            dataset.measurementType = to_measurement_type(line)
        elif i == lastLineOfGeneralInfo:
            break

    return linesIt


def get_values_from(line: str) -> List[str]:
    line = line.rstrip()
    return line.split()


def get_key_and_values_from(line: str) -> Tuple[str, List[str]]:
    line = line.rstrip()
    assert line.count("=") == 1
    key, valuesAsSingleString = line.split("=", maxsplit=2)
    key = key.strip()
    values = get_values_from(valuesAsSingleString)

    return key, values


def read_number_of_model_parameters_and_variable_switching_points_to_dataset(values: List[str], linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    assert len(values) >= 2

    dataset.numberOfModelParameters = int(values[0])
    dataset.numberOfVariableSwitchingPoints = int(values[1])

    return linesIt


def read_constraints_info_to_dataset(values: List[str], linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    assert len(values) == 1

    dataset.totalNumberOfConstraints = int(values[0])

    key, values = get_key_and_values_from(next(linesIt))
    assert key == "NEQU"
    assert len(values) == 1
    dataset.numberOfEqualityConstraints = int(values[0])

    constraintsInfos = []
    for _ in range(dataset.totalNumberOfConstraints):
        values = get_values_from(next(linesIt))
        assert len(values) == 3
        constraintsInfos.append(ConstraintInfo(constraintTime=float(values[1]), constraintConcentration=float(values[2])))

    dataset.constraintsInfos = constraintsInfos

    return linesIt


def read_number_of_odes_to_dataset(values: List[str], linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    assert len(values) == 1
    dataset.numberOfOdes = int(values[0])
    return linesIt


def read_number_of_concentrations_to_dataset(values: List[str], linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    assert len(values) == 1
    dataset.numberOfConcentrations = int(values[0])
    return linesIt


def read_number_of_time_points_and_logarithmic_scaling_of_x_axis_to_dataset(values: List[str], linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    assert len(values) == 2
    dataset.numberOfTimePoints = int(values[0])
    dataset.useLogarithmicScalingOfXAxis = bool(int(values[1]))
    return linesIt


def read_dimension_of_fitting_function_to_dataset(values: List[str], linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    assert len(values) == 1
    dataset.dimensionOfFittingFunction = int(values[0])
    return linesIt


def read_dimensions_to_dataset(linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    parserLookup = {
        "NPAR": read_number_of_model_parameters_and_variable_switching_points_to_dataset,
        "NRES": read_constraints_info_to_dataset,
        "NODE": read_number_of_odes_to_dataset,
        "NCONC": read_number_of_concentrations_to_dataset,
        "NTIME": read_number_of_time_points_and_logarithmic_scaling_of_x_axis_to_dataset,
        "NMEAS": read_dimension_of_fitting_function_to_dataset}

    lastKeyOfDimensionSection = "NMEAS"

    line = next(linesIt, None)
    while line:
        key, values = get_key_and_values_from(line)

        if not key in parserLookup.keys():
            logging.warning(f"Unexpected key {key}. Skip line {line}.")
            line = next(linesIt, None)
            continue
        else:
            linesIt = parserLookup[key](values, linesIt, dataset)

        if key == lastKeyOfDimensionSection:
            break
        else:
            line = next(linesIt, None)

    return linesIt


def skip_lines_with_no_useful_information(linesIt: Iterator[str]) -> Iterator[str]:
    keysToSkip = ["NPLOT", "NOUT", "METHOD", "OPTP1", "OPTP2", "OPTP3", "OPTE1",
                  "OPTE2", "OPTE3", "ODEP1", "ODEP2", "ODEP3", "ODEP4", "ODEE1", "ODEE2"]
    lastKeyBeforeBoundsSection = "ODEE3"

    line = next(linesIt, None)
    while line:
        key, _ = get_key_and_values_from(line)

        if key == lastKeyBeforeBoundsSection:
            break

        if key in keysToSkip:
            # skip line
            line = next(linesIt, None)
            continue
        else:
            logging.warning(f"Found key '{key}' while skipping lines that that is not in the list of keys to skip")
            break

    return linesIt


def read_model_parameter_infos_to_dataset(linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    for _ in range(dataset.numberOfModelParameters):
        values = get_values_from(next(linesIt))
        assert len(values) == 4

        name = values[0]
        lowerBound = float(values[1])
        initialGuess = float(values[2])
        upperBound = float(values[3])

        dataset.modelParameterInfos.append(ModelParameterInfo(
            name=name,
            initialGuess=initialGuess,
            lowerBound=lowerBound,
            upperBound=upperBound))

    return linesIt


def read_scaling_information_to_dataset(linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    line = next(linesIt)
    key, values = get_key_and_values_from(line)

    assert key == "SCALE"
    assert len(values) == 1

    dataset.scalingInformation = int(values[0])

    return linesIt


def read_measurements_to_dataset(linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    hasConcentration = (dataset.numberOfConcentrations > 0)

    if hasConcentration:
        numberOfConcentrationEntries = 1
    else:
        numberOfConcentrationEntries = 0

    numberOfTimeEntries = 1
    numberOfMeasurements = dataset.dimensionOfFittingFunction
    numberOfWeights = dataset.dimensionOfFittingFunction

    expectedNumberOfMeasurementEntriesInRow = numberOfTimeEntries + \
        numberOfConcentrationEntries + numberOfMeasurements + numberOfWeights

    for _ in range(max(1, dataset.numberOfConcentrations)):
        concentrationValue = None
        timePoints = []
        measurements = []
        weights = []
        for _ in range(dataset.numberOfTimePoints):
            values = [float(val) for val in get_values_from(next(linesIt))]

            assert len(values) == expectedNumberOfMeasurementEntriesInRow

            timePoints.append(values[0])

            if hasConcentration:
                if concentrationValue is not None:
                    if concentrationValue != values[1]:
                        raise MeasurementInconsistencyError(f"{dataset.name}: Measurements dimensions inconsistent")
                else:
                    concentrationValue = values[1]

            measurementRow = []
            weightsRow = []
            measurementsStart = numberOfTimeEntries + numberOfConcentrationEntries
            for i, val in enumerate(values[measurementsStart:]):
                if i % 2 == 0:
                    measurementRow.append(val)
                else:
                    weightsRow.append(val)

            measurements.append(measurementRow)
            weights.append(weightsRow)

        dataset.measurements.append(Measurements(concentrationValue, timePoints, measurements, weights))

    return linesIt


def skip_lines_with_auxiliary_parameters(linesIt: Iterator[str]) -> Iterator[str]:
    numberOfAuxParams = 3
    for i in range(numberOfAuxParams):
        next(linesIt)

    return linesIt


def read_constant_switching_points_to_dataset(linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    key, values = get_key_and_values_from(next(linesIt))
    assert key == "NDISCO"
    assert len(values) == 1
    dataset.numberOfConstantSwitchingPoints = int(values[0])

    for i, line in enumerate(linesIt):
        if i == dataset.numberOfConstantSwitchingPoints:
            break

        values = get_values_from(line)
        assert len(values) == 1
        dataset.constantSwitchingPoints.append(float(values[0]))

    return linesIt


def to_float(val: str) -> float:
    val = val.replace(",", ".")
    val = val.replace('"', "")
    return float(val)


def read_exact_model_parameters_to_dataset(linesIt: Iterator[str], dataset: Dataset) -> Iterator[str]:
    line = next(linesIt, None)
    while line:
        if "Exact solution" in line:
            exactValues = []
            for i, line in enumerate(linesIt):
                if i == dataset.numberOfModelParameters:
                    break

                values = get_values_from(line)

                if len(values) != 1:
                    logging.warning(f"{dataset.name}: Failed to read exact model parameters")
                    break
                exactValues.append(to_float(values[0]))

            if len(exactValues) == dataset.numberOfModelParameters:
                for i, val in enumerate(exactValues):
                    dataset.modelParameterInfos[i].exactValue = val

        line = next(linesIt, None)

    return linesIt


def read_dataset_from_file_object(file: TextIO) -> Dataset:
    dataset = Dataset()
    linesIt = iter(file.readlines())

    linesIt = read_general_info_to_dataset(linesIt, dataset)

    if dataset.modelType in [ModelType.PartialDifferentialEquations, ModelType.PartialDifferentialAlgebraicEquations]:
        logging.info(f"Skip model {dataset.name}, PDE models are not supported")
        return dataset

    linesIt = read_dimensions_to_dataset(linesIt, dataset)
    linesIt = skip_lines_with_no_useful_information(linesIt)
    linesIt = read_model_parameter_infos_to_dataset(linesIt, dataset)
    linesIt = read_scaling_information_to_dataset(linesIt, dataset)
    linesIt = read_measurements_to_dataset(linesIt, dataset)
    linesIt = skip_lines_with_auxiliary_parameters(linesIt)
    linesIt = read_constant_switching_points_to_dataset(linesIt, dataset)
    linesIt = read_exact_model_parameters_to_dataset(linesIt, dataset)

    return dataset


def read_dataset_from_file(fileName: os.path) -> Dataset:
    assert os.path.isfile(fileName)
    logging.info(f"Read dataset {fileName}")

    with open(fileName, encoding='utf-8', errors='ignore') as file:
        return read_dataset_from_file_object(file)


def read_datasets_from_directory(directory: str) -> List[Dataset]:
    datasets = []
    for fileName in os.listdir(directory):
        fileName = os.path.join(directory, fileName)
        if os.path.isfile(fileName) and fileName.endswith(".DAT"):
            try:
                datasets.append(read_dataset_from_file(fileName))
            except MeasurementInconsistencyError as error:
                logging.error(error)

    return datasets
