from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional
import re
import logging

class VariablesDimensionError(Exception):
    def __init__(self, expectedNumberOfVariables: int, observedNumberOfVariables: int):
        self.message = f"Wrong number of variables detected, expected: {expectedNumberOfVariables}, observed: {observedNumberOfVariables}."
        super().__init__(self.message)


class FunctionsDimensionError(Exception):
    def __init__(self, expectedNumberOfFunctions: int, observedNumberOfFunctions: int):
        self.message = f"Wrong number of functions detected, expected: {expectedNumberOfFunctions}, observed: {observedNumberOfFunctions}."
        super().__init__(self.message)


@dataclass
class PCOMPConstant:
    name: str
    value: float 
   

@dataclass
class PCOMPVariables:
    modelParameters: list[str] = field(default_factory=list)
    states: list[str] = field(default_factory=list)
    concentration: Optional[str] = None
    time: str = ""


@dataclass
class Function:
    name: str = ""
    body: list[str] = field(default_factory=list)


@dataclass
class PCOMPFunctions:
    differentialEquation: list[Function] = field(default_factory=list)
    initialValues: list[Function] = field(default_factory=list)
    fittingFunction: list[Function] = field(default_factory=list)
    constraints: Optional[list[Function]] = None


@dataclass
class PCOMPModel:
    constants: list[PCOMPConstant]
    variables: PCOMPVariables
    functions: PCOMPFunctions
    

@dataclass
class PCOMPModelDimension:
    numberOfModelParameters: int
    numberOfOdes: int
    numberOfConcentrations: int
    numberOfFittingFunctions: int
    numberOfConstraints: int


def remove_comments(modelFileContent: str) -> str:
    modelFileLines = modelFileContent.splitlines()
    modelFileLinesWithoutComments = [line for line in modelFileLines if not line.startswith(("c", "C"))]
    return '\n'.join(modelFileLinesWithoutComments)


def get_pcomp_special_line_pattern() -> re.Pattern:
    return re.compile(r"^\*", flags=re.IGNORECASE|re.MULTILINE)


def get_pcomp_function_pattern() -> re.Pattern:
    return re.compile(r"^(\*\s*FUNCTION)", flags=re.IGNORECASE|re.MULTILINE)


def read_pcomp_block(modelFileContent: str,  blockStartIndex: int) -> str:
    findNewSpecialLinePattern = get_pcomp_special_line_pattern()

    blockEndMatch = findNewSpecialLinePattern.search(modelFileContent[blockStartIndex + 1:])
    if blockEndMatch:
        blockEndIndex = blockStartIndex + blockEndMatch.end()
    else:
        blockEndIndex = -1

    return modelFileContent[blockStartIndex:blockEndIndex]


def split_variables(variablesAsList: list[str], dimension: PCOMPModelDimension) -> PCOMPVariables:
    numberOfTimeVariables = 1 
    hasConcentration = dimension.numberOfConcentrations > 0
    numberOfConcentrationVariables = 1 if hasConcentration else 0
    expectedNumberOfVariables = dimension.numberOfModelParameters + dimension.numberOfOdes + numberOfConcentrationVariables + numberOfTimeVariables
    
    if len(variablesAsList) != expectedNumberOfVariables:
        raise VariablesDimensionError(expectedNumberOfVariables, len(variablesAsList))

    variables = PCOMPVariables()

    modelParametersBegin = 0
    modelParametersEnd = dimension.numberOfModelParameters
    variables.modelParameters = variablesAsList[modelParametersBegin:modelParametersEnd]

    statesBegin = modelParametersEnd
    statesEnd = statesBegin + dimension.numberOfOdes
    variables.states = variablesAsList[statesBegin:statesEnd]

    if hasConcentration:
        variables.concentration = variablesAsList[-2]
    
    variables.time = variablesAsList[-1]

    return variables


def to_function(functionAsString: str) -> Function:
    findNewFunctionPattern = get_pcomp_function_pattern()

    splittedFunction = functionAsString.splitlines()

    functionName = findNewFunctionPattern.sub("", splittedFunction[0]).strip()
    functionBody: list[str] = [line.strip() for line in splittedFunction[1:] if line.strip()]
    
    return Function(name=functionName, body=functionBody)


def split_functions(functionsAsList: list[str], dimension: PCOMPModelDimension) -> PCOMPFunctions:
    expectedNumberOfFunctions = 2 * dimension.numberOfOdes + dimension.numberOfFittingFunctions + dimension.numberOfConstraints

    if len(functionsAsList) != expectedNumberOfFunctions:
        raise FunctionsDimensionError(expectedNumberOfFunctions, len(functionsAsList))

    functions = PCOMPFunctions()

    differentialEquationBegin = 0
    differentialEquationEnd = dimension.numberOfOdes
    functions.differentialEquation = [to_function(f) for f in functionsAsList[differentialEquationBegin:differentialEquationEnd]]

    initialValuesBegin = differentialEquationEnd
    initialValuesEnd = initialValuesBegin + dimension.numberOfOdes
    functions.initialValues = [to_function(f) for f in functionsAsList[initialValuesBegin:initialValuesEnd]]

    fittingFunctionBegin = initialValuesEnd
    fittingFunctionEnd = fittingFunctionBegin + dimension.numberOfFittingFunctions
    functions.fittingFunction = [to_function(f) for f in functionsAsList[fittingFunctionBegin:fittingFunctionEnd]]

    if dimension.numberOfConstraints > 0:
        functions.constraints = [to_function(f) for f in functionsAsList[fittingFunctionEnd:]]

    return functions


def get_pcomp_model_dimensions_from(dataset) -> PCOMPModelDimension:
    return PCOMPModelDimension(numberOfModelParameters=dataset.numberOfModelParameters,
                               numberOfOdes=dataset.numberOfOdes,
                               numberOfConcentrations=dataset.numberOfConcentrations,
                               numberOfFittingFunctions=dataset.dimensionOfFittingFunction,
                               numberOfConstraints=dataset.totalNumberOfConstraints)


def read_pcomp_model(modelFile: Path, dimension: PCOMPModelDimension) -> PCOMPModel:
    modelFileContent = modelFile.read_text()
    modelFileContent = remove_comments(modelFileContent)

    variables: list[str] = []
    functions: list[str] = []

    indexSets: list[tuple[str, str]] = []

    realConstants: list[PCOMPConstant] = []

    findNewSpecialLinePattern = get_pcomp_special_line_pattern()
    findNewSetOfIndicesPattern = re.compile(r"^(\*\s*SET OF INDICES)", flags=re.IGNORECASE|re.MULTILINE)
    findNewRealConstantsPattern = re.compile(r"^(\*\s*REAL CONSTANT[S]?)", flags=re.IGNORECASE|re.MULTILINE)
    findNewVariablePattern = re.compile(r"^(\*\s*VARIABLE)", flags=re.IGNORECASE|re.MULTILINE)
    findNewFunctionPattern = get_pcomp_function_pattern()
        
    for specialLineStartMatch in findNewSpecialLinePattern.finditer(modelFileContent):
        if findNewSetOfIndicesPattern.match(modelFileContent[specialLineStartMatch.start():]):
            indexSetBlock = read_pcomp_block(modelFileContent, specialLineStartMatch.start())
            indexSetBlock = findNewSetOfIndicesPattern.sub("", indexSetBlock)
            indexSetBlock = indexSetBlock.strip()
            indexSetBlock = indexSetBlock.strip('\n')
            indexSetBlock = indexSetBlock.strip()
            for line in indexSetBlock.splitlines():
                splitIndexSet = line.split('=', maxsplit=1)
                assert len(splitIndexSet) == 2
                indexSets.append((splitIndexSet[0].strip(), splitIndexSet[1].strip()))
            continue

        if findNewRealConstantsPattern.match(modelFileContent[specialLineStartMatch.start():]):
            realConstantsBlock = read_pcomp_block(modelFileContent, specialLineStartMatch.start())
            realConstantsBlock = findNewRealConstantsPattern.sub("", realConstantsBlock)
            realConstantsBlock = realConstantsBlock.strip()
            realConstantsBlock = realConstantsBlock.strip('\n')
            realConstantsBlock = realConstantsBlock.strip()
            for line in realConstantsBlock.splitlines():
                splitRealConstant = line.split('=', maxsplit=1)
                assert len(splitRealConstant) == 2
                constantName = splitRealConstant[0].strip()
                constantValueAsStr = splitRealConstant[1].strip()
                try:
                    constantValue = float(constantValueAsStr)
                    realConstants.append(PCOMPConstant(constantName, constantValue))
                except ValueError as err:
                    logging.warning(f"Unable to read value of constant '{constantName}'\nObserved value: '{constantValueAsStr}'\nError: {err}\nSkip constant")
            continue

        if findNewFunctionPattern.match(modelFileContent[specialLineStartMatch.start():]):
            functions.append(read_pcomp_block(modelFileContent, specialLineStartMatch.start()))
            continue

        if findNewVariablePattern.match(modelFileContent[specialLineStartMatch.start():]):
            variablesBlock = read_pcomp_block(modelFileContent, specialLineStartMatch.start())
            variablesBlock = findNewVariablePattern.sub("", variablesBlock)
            for line in variablesBlock.splitlines():
                if any([indexSet[0] in line for indexSet in indexSets]):
                    variables.append(line.strip())
                else:
                    variables.extend([var for var in line.replace(' ', '').split(',') if var])
            
    return PCOMPModel(realConstants, split_variables(variables, dimension), split_functions(functions, dimension))

