import logging
import re
import os
from typing import Dict, Callable, List, Optional
from collections.abc import Iterable

from .Dataset import Dataset, ConstraintInfo, ModelParameterInfo, ModelType


def from_dict(dict: Dict) -> Callable[[re.Match], str]:
    def lookup(match: re.Match) -> str:
        key = match.group(1)
        if not key in dict:
            raise ValueError(f"Lookup dict does not contain key '{key}'")
        return dict[key]
    return lookup


def to_cpp_bool_name(value: bool) -> str:
    if value:
        return "true"
    else:
        return "false"


def to_cpp_array_of_constraints_infos(constraintsInfos: List[ConstraintInfo]) -> str:
    entries = ", ".join([f"ConstraintInfo{{ {c.constraintTime}, {c.constraintTime} }}" for c in constraintsInfos])
    return f"{{ {entries} }}"


def to_cpp_optional(value: Optional[float]) -> str:
    if value is None:
        return "{}"

    return f"{value}"


def to_cpp_array_of_model_parameter_infos(modelParameterInfos: List[ModelParameterInfo]) -> str:
    entries = ", ".join(
        [f'ModelParameterInfo{{ "{m.name}", {m.initialGuess}, {m.lowerBound}, {m.upperBound}, {to_cpp_optional(m.exactValue)} }}' for m in modelParameterInfos])
    return f"{{ {entries} }}"


def to_cpp_initializer_list(values: Iterable[float]) -> str:
    entries = ", ".join([f"{v}" for v in values])
    return f"{{ {entries} }}"


def write_to_cpp_header(subs: dict[str, str], templateFilePath: os.path, outputFilePath: os.path) -> None:
    logging.info(f"Use template file {templateFilePath}")
    assert os.path.isfile(templateFilePath)

    if os.path.exists(outputFilePath):
        assert os.path.isfile(outputFilePath)
        logging.warning(f"File {outputFilePath} exists and will be overridden")

    logging.info(f"Replace keywords in {templateFilePath} and write result to {outputFilePath}")

    with open(templateFilePath, mode='r', encoding='utf-8') as templateFile, open(outputFilePath, mode='w', encoding='utf-8') as outputFile:
        for line in templateFile:
            logging.debug(f"read line '{line}'")
            newLine = re.sub('@(.*?)@', from_dict(subs), line)
            logging.debug(f"write line: '{newLine}'")
            outputFile.write(newLine)


def write_dataset_to_cpp_header(dataset: Dataset, templateFilePath: os.path, outputFilePath: os.path) -> None:
    assert dataset.modelType == ModelType.OrdinaryDifferentialEquations

    subs = {
        "name": f'"{dataset.name}"',
        "description": f'"{dataset.description}"',
        "measurementType": f"MeasurementType::{dataset.measurementType.name}",
        "numberOfModelParameters": f"{dataset.numberOfModelParameters}",
        "numberOfOdes": f"{dataset.numberOfOdes}",
        "dimensionOfFittingFunction": f"{dataset.dimensionOfFittingFunction}",
        "totalNumberOfConstraints": f"{dataset.totalNumberOfConstraints}",
        "numberOfEqualityConstraints": f"{dataset.numberOfEqualityConstraints}",
        "numberOfTimePoints": f"{dataset.numberOfTimePoints}",
        "numberOfConcentrations": f"{dataset.numberOfConcentrations}",
        "numberOfVariableSwitchingPoints": f"{dataset.numberOfVariableSwitchingPoints}",
        "numberOfConstantSwitchingPoints": f"{dataset.numberOfConstantSwitchingPoints}",
        "useLogarithmicScalingOfXAxis": to_cpp_bool_name(dataset.useLogarithmicScalingOfXAxis),
        "scalingInformation": f"{dataset.scalingInformation}",
        "constraintsInfos": to_cpp_array_of_constraints_infos(dataset.constraintsInfos),
        "modelParameterInfos": to_cpp_array_of_model_parameter_infos(dataset.modelParameterInfos),
        "constantSwitchingPoints": to_cpp_initializer_list(dataset.constantSwitchingPoints),
        "timeRange": to_cpp_initializer_list(dataset.get_time_range()),
        "concentrations": to_cpp_initializer_list(dataset.get_concentration_values())
    }

    write_to_cpp_header(subs, templateFilePath, outputFilePath)
