import logging
import os
from typing import Optional, List

import numpy as np

from .Dataset import Measurements


def check_path(outputFilePath: os.path) -> None:
    if os.path.exists(outputFilePath):
        assert os.path.isfile(outputFilePath)
        logging.warning(f"File {outputFilePath} exists and will be overridden")


def write_measurement_values_to_csv_file(
        timePoints: List[float],
        measurementValues: List[List[float]],
        concentrationValue: Optional[float],
        outputMeasurementFilePath: os.path) -> None:
    check_path(outputMeasurementFilePath)
    timePoints = np.asarray(timePoints)
    timePoints = np.expand_dims(timePoints, axis=1)
    measurementValues = np.asarray(measurementValues)

    header = ["time"] + [f"measurement_{i}" for i in range(measurementValues.shape[1])]

    if concentrationValue is None:
        concentrationValue = np.empty((len(timePoints), 0))
    else:
        concentrationValue = concentrationValue * np.ones_like(timePoints)
        header.append("concentration")

    stackedMeasurements = np.hstack((timePoints, measurementValues, concentrationValue))
    delimiter = ','
    np.savetxt(outputMeasurementFilePath, stackedMeasurements, header=delimiter.join(header), delimiter=delimiter, comments='')


def write_weights_to_csv_file(
        weights: List[List[float]],
        outputWeightsFilePath: os.path) -> None:
    check_path(outputWeightsFilePath)

    weights = np.asfarray(weights)
    header = [f"measurement_{i}" for i in range(weights.shape[1])]
    
    delimiter = ','
    np.savetxt(outputWeightsFilePath, weights, header=delimiter.join(header), delimiter=delimiter, comments='')


def write_measurements_to_csv_files(measurements: Measurements, outputMeasurementFilePath: os.path, outputWeightsFilePath: os.path) -> None:
    write_measurement_values_to_csv_file(
        measurements.timePoints,
        measurements.measurements,
        measurements.concentrationValue,
        outputMeasurementFilePath)

    write_weights_to_csv_file(measurements.weights, outputWeightsFilePath)
