#pragma once

#include "dynamic_model_c_interface.h"
#include "dynamic_model_infos.hpp"
#include "InvalidFunctionCallException.hpp"

#include <gsl/gsl_assert>

#include <algorithm>
#include <array>
#include <iterator>
#include <span>

namespace pi_database::dynamic_model
{

namespace internal
{

constexpr auto getNumberOfVariables()
{
  constexpr int numberOfTimeEntries = 1;
  if constexpr (getNumberOfConcentrations() > 0) {
    constexpr auto numberOfConcentrationEntries = 1;
    return getNumberOfModelParameters()
           + getNumberOfOdes()
           + numberOfConcentrationEntries
           + numberOfTimeEntries;
  } else {
    return getNumberOfModelParameters()
           + getNumberOfOdes()
           + numberOfTimeEntries;
  }
}

constexpr auto getNumberOfFunctions()
{
  constexpr auto numberOfInitialValues = getNumberOfOdes();
  return getNumberOfOdes()
         + numberOfInitialValues
         + getDimensionOfFittingFunction()
         + getTotalNumberOfConstraints();
}

constexpr void checkInputDimensions(
  const std::span<const double> states,
  const std::span<const double> modelParameters)
{
  Expects(std::size(states) == getNumberOfOdes());
  Expects(std::size(modelParameters) == getNumberOfModelParameters());
}

constexpr auto generateInputWithModelParametersAndStates(
  const std::span<const double> modelParameters,
  const std::span<const double> states)
{
  std::array<double, getNumberOfVariables()> input{};
  auto* beginStates = std::copy(
    std::cbegin(modelParameters),
    std::cend(modelParameters),
    std::begin(input));

  std::copy(
    std::cbegin(states),
    std::cend(states),
    beginStates);

  return input;
}

constexpr auto generateInput(
  const double time,
  const std::span<const double> states,
  const double concentration,
  const std::span<const double> modelParameters)
{
  auto input = generateInputWithModelParametersAndStates(modelParameters, states);
  input[getNumberOfModelParameters() + getNumberOfOdes()] = concentration;
  // Marek (30.05.2022): This internal function is only called when getNumberOfConcentrations() > 0 which is expected
  // in the calling function. In that case, the array size is sufficient and there is no out-of-bound access.
  input[getNumberOfModelParameters() + getNumberOfOdes() + 1U] = time;// NOLINT [cppcoreguidelines-pro-bounds-constant-array-index]

  return input;
}

constexpr auto generateInput(
  const double time,
  const std::span<const double> states,
  const std::span<const double> modelParameters)
{
  auto input = generateInputWithModelParametersAndStates(modelParameters, states);
  input[getNumberOfModelParameters() + getNumberOfOdes()] = time;

  return input;
}

template<std::size_t TargetFunctionSize, std::ptrdiff_t Offset>
constexpr auto isActive()
{
  std::array<bool, getNumberOfFunctions()> isActive{};
  isActive.fill(false);
  std::fill_n(
    std::next(std::begin(isActive), Offset),
    TargetFunctionSize,
    true);

  return isActive;
}

template<std::size_t TargetFunctionSize, std::ptrdiff_t Offset>
auto call_xfun_with_input(const std::array<double, getNumberOfVariables()>& input)
{
  constexpr auto targetFunctionIsActive = isActive<TargetFunctionSize, Offset>();

  std::array<double, getNumberOfFunctions()> output{};

  const auto errorCode = xfun_wrapper(
    input.data(),
    getNumberOfVariables(),
    output.data(),
    getNumberOfFunctions(),
    targetFunctionIsActive.data());

  if (errorCode > 0) {
    throw InvalidFunctionCallException{ errorCode };
  }

  std::array<double, TargetFunctionSize> targetFunctionValues{};
  std::copy(
    std::next(std::begin(output), Offset),
    std::next(std::begin(output), Offset + TargetFunctionSize),
    std::begin(targetFunctionValues));

  return targetFunctionValues;
}

}// namespace internal

template<std::size_t TargetFunctionSize, std::ptrdiff_t Offset>
auto call_xfun(
  const double time,
  const std::span<const double> states,
  const std::span<const double> modelParameters)
{
  Expects(getNumberOfConcentrations() == 0);
  internal::checkInputDimensions(states, modelParameters);

  const auto input = internal::generateInput(time, states, modelParameters);
  return internal::call_xfun_with_input<TargetFunctionSize, Offset>(input);
}

template<std::size_t TargetFunctionSize, std::ptrdiff_t Offset>
auto call_xfun(
  const double time,
  const std::span<const double> states,
  const double concentration,
  const std::span<const double> modelParameters)
{
  Expects(getNumberOfConcentrations() > 0);
  internal::checkInputDimensions(states, modelParameters);

  const auto input = internal::generateInput(time, states, concentration, modelParameters);
  return internal::call_xfun_with_input<TargetFunctionSize, Offset>(input);
}

}// namespace pi_database::dynamic_model
